import React, { useState } from 'react';
import { useMarket } from '../../utils/markets';
import DataTable from '../layout/DataTable';

import styled from 'styled-components';
import { Button, Row, Col, Tag } from 'antd';
import { cancelOrder, cancelOrders } from '../../utils/send';
import { useWallet } from '../../utils/wallet';
import { useSendConnection } from '../../utils/connection';
import { notify } from '../../utils/notifications';
import { DeleteOutlined } from '@ant-design/icons';

const CancelButton = styled(Button)`
  color: #f23b69;
  border: 1px solid #f23b69;
`;

export default function OpenOrderTable({ openOrders }) {
  let { market } = useMarket();
  let { wallet } = useWallet();
  let connection = useSendConnection();

  const [cancelAllInProgress, setCancelAllInProgress] = useState(false);
  const [cancelId, setCancelId] = useState(null);

  async function cancel(order) {
    try {
      await cancelOrder({
        order,
        market,
        connection,
        wallet,
        onBeforeSendCallBack: () => setCancelId(order?.orderId),
        onConfirmCallBack: () => setCancelId(null),
      });
    } catch (e) {
      notify({
        message: 'Error cancelling order: ' + e.message,
        type: 'error',
      });
      setCancelId(null);
    }
  }

  async function cancelAll() {
    try {
      await cancelOrders({
        orders: openOrders,
        market,
        connection,
        wallet,
        onBeforeSendCallBack: () => setCancelAllInProgress(true),
        onConfirmCallBack: () => setCancelAllInProgress(false),
      });
    } catch (e) {
      notify({
        message: 'Error cancelling orders: ' + e.message,
        type: 'error',
      });
      setCancelAllInProgress(false);
    }
  }

  const columns = [
    {
      title: 'Market',
      dataIndex: 'marketName',
      key: 'marketName',
    },
    {
      title: 'Side',
      dataIndex: 'side',
      key: 'side',
      render: (side) => (
        <Tag
          color={side === 'buy' ? '#41C77A' : '#F23B69'}
          style={{ fontWeight: 700 }}
        >
          {side.charAt(0).toUpperCase() + side.slice(1)}
        </Tag>
      ),
    },
    {
      title: 'Size',
      dataIndex: 'size',
      key: 'size',
    },
    {
      title: 'Price',
      dataIndex: 'price',
      key: 'price',
    },
    {
      key: 'orderId',
      render: (order) => (
        <div style={{ textAlign: 'right' }}>
          <CancelButton
            icon={<DeleteOutlined />}
            onClick={() => cancel(order)}
            loading={cancelId + '' === order?.orderId + ''}
          >
            Cancel
          </CancelButton>
        </div>
      ),
    },
  ];
  const dataSource = (openOrders || []).map((order) =>
    Object.assign(order, { key: order.orderId }),
  );

  return (
    <>
      {openOrders?.length > 0 && (
        <Row justify="end">
          <Col style={{ paddingRight: 15 }}>
            <CancelButton
              icon={<DeleteOutlined />}
              onClick={cancelAll}
              loading={cancelAllInProgress}
            >
              Cancel all
            </CancelButton>
          </Col>
        </Row>
      )}
      <Row>
        <Col span={24}>
          <DataTable
            emptyLabel="No open orders"
            dataSource={dataSource}
            columns={columns}
            pagination={true}
            pageSize={5}
          />
        </Col>
      </Row>
    </>
  );
}
